
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: CreateFrame, CreateFromMixins, Mixin, next, NORMAL_FONT_COLOR, SixUI, string, table, type, unpack, UIParent

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local MAJOR, MINOR = "SixTooltip-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") end
local Library = LibStub:NewLibrary(MAJOR, MINOR)
if not Library then return end
local LibQTip = LibStub("LibQTip-1.0")

local CONFIG_DEFAULTS = {
	anchor = nil,
	autoHideAnchor = nil,
	cellMarginH = 6,
	cellMarginV = 3,
	columnJustification = {"LEFT"},
	enableScrolling = true,
	enableMouse = true,
	font = GameTooltipText,
	frameStrata = "DIALOG",
	headerFont = GameTooltipHeaderText,
	ignoreReleaseAllTooltips = false,
	key = "",
	numColumns = 1,
	point = nil,
	separatorHeight = 1,
	skipReleaseAllTooltips = false,
}

-------------------------------------------------------------------------------
-- LibQTip
-------------------------------------------------------------------------------

-- Release all tooltips other than excludeKey.
local function ReleaseAllTooltips(excludeKey)
	for key, tooltip in LibQTip:IterateTooltips() do
		if key ~= excludeKey and string.sub(key or "", 1 ,1) ~= "!" then
			tooltip:Release()
		end
	end
end

-- Configure an interactive tooltip.
local function ConfigureTooltip(tooltip, config)
	tooltip:EnableMouse(config.enableMouse)
	tooltip:SetFrameStrata(config.frameStrata)

	if config.anchor then
		tooltip:SetAutoHideDelay(0.1, config.autoHideAnchor or config.anchor)
		tooltip:SmartAnchorTo(config.anchor)
	elseif config.point then
		if config.autoHideAnchor then
			tooltip:SetAutoHideDelay(0.1, config.autoHideAnchor)
		else
			tooltip:SetAutoHideDelay(0)
		end

		tooltip:ClearAllPoints()
		tooltip:SetPoint(unpack(config.point))
	end
end

-- Apply SixUI's style to a tooltip.
local function StyleTooltip(tooltip, config)
	if SixUI then
		SixUI[2]:CreateBorder(tooltip)
		tooltip:SetBackdropBorderColor(0, 0, 0, 0)
		tooltip:SetBackdropColor(0, 0, 0, 0)
		tooltip:SetFont(config.font)
		tooltip:SetHeaderFont(config.headerFont)
	end
end

-- Set up a tooltip for SixUI.
local function AcquireTooltip(tooltip, config)
	if not config.skipReleaseAllTooltips then
		ReleaseAllTooltips(config.key)
	end

	if not tooltip or not LibQTip:IsAcquired(config.key) then
		tooltip = LibQTip:Acquire(config.key, config.numColumns, unpack(config.columnJustification))
		ConfigureTooltip(tooltip, config)
		StyleTooltip(tooltip, config)
	end

	tooltip:Clear()
	tooltip:SetCellMarginH(config.cellMarginH)
	tooltip:SetCellMarginV(config.cellMarginV)

	return tooltip
end

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local SixTooltipMixin = {}

local function InjectDynamicLines(self)
	local width = self.tooltip:GetWidth() - 24

	for _, line in next, self.dynamic do
		self.tooltip:SetCell(line.lineNum, line.colNum, line.value, line.font, line.justification, line.colSpan, nil, nil, nil, width)
	end

	self.dynamic = {}
end

function SixTooltipMixin:AddDynamicLine(label, func, arg, isHeader)
	return self:AddSpannedLine(label, func, arg, isHeader, true)
end

function SixTooltipMixin:AddSpannedHeader(label, func, arg, skipSeparator)
	return self:AddSpannedLine(label, func, arg, true, false, skipSeparator)
end

function SixTooltipMixin:SetText(text, _, _, _, isDynamic)
	return self:AddSpannedLine(text, nil, nil, isDynamic, false)
end

function SixTooltipMixin:AddSpannedLine(label, func, arg, isHeader, isDynamic, skipSeparator)
	local line = isHeader and self.tooltip:AddHeader("") or self.tooltip:AddLine("")

	if isDynamic then
		table.insert(self.dynamic, {
			lineNum = line,
			colNum = 1,
			value = label,
			font = nil,
			justification = "LEFT",
			colSpan = self.config.numColumns,
		})
	else
		self.tooltip:SetCell(line, 1, label, nil, "LEFT", self.config.numColumns)
	end

	if type(func) == "function" then
		self.tooltip:SetLineScript(line, "OnMouseUp", func, arg)
	end

	if type(isHeader) == "boolean" and isHeader and not skipSeparator then
		self.tooltip:AddSeparator(self.config.separatorHeight)
	end

	return line
end

function SixTooltipMixin:IsShown()
	return self.tooltip and LibQTip:IsAcquired(self.config.key)
end

function SixTooltipMixin:Refresh()
	if self:IsShown() then
		if type(self.OnRefresh) == "function" then
			self:OnRefresh()
		end

		self:Show()
	end
end

function SixTooltipMixin:Show(...)
	self.tooltip = AcquireTooltip(self.tooltip, self.config)

	if type(self.OnShow) == "function" then
		if self:OnShow(...) then
			InjectDynamicLines(self)
		end
	end

	for colNum = 1, self.config.numColumns do
		self.tooltip:SetColumnTextColor(colNum, NORMAL_FONT_COLOR:GetRGB())
	end

	if self.config.enableScrolling then
		self.tooltip:UpdateScrolling()
	end

	self.tooltip:Show()
end

function SixTooltipMixin:SmartAnchorTo(anchor, offsetX, offsetY)
	self.tooltip:ClearAllPoints()

	if anchor:GetLeft() > UIParent:GetWidth() / 2 then
		self.tooltip:SetPoint("RIGHT", anchor, "LEFT", -offsetX, offsetY)
	else
		self.tooltip:SetPoint("LEFT", anchor, "RIGHT", offsetX, offsetY)
	end

	if anchor:GetTop() > UIParent:GetHeight() / 2 then
		self.tooltip:SetPoint("TOP", anchor, "LEFT", -offsetX, offsetY)
	else
		self.tooltip:SetPoint("BOTTOM", anchor, "RIGHT", offsetX, offsetY)
	end
end

function SixTooltipMixin:UpdateConfig(config)
	self.config = Mixin(self.config, config)

	if self:IsShown() then
		ConfigureTooltip(self.tooltip, self.config)
		StyleTooltip(self.tooltip, self.config)
	end
end

local function CreateInstance(config)
	local SixTooltip = CreateFrame("Frame")

	-- Wrapper
	for _, key in next, {"AcquireCell", "AddColumn", "AddHeader", "AddLine", "AddSeparator", "Clear", "GetCellPrototype", "GetColumnCount", "getContentHeight", "GetDefaultProvider", "GetFont", "GetHeaderFont", "GetLineCount", "GetPosition", "Hide", "HookScript", "InitializeCell", "IsAcquiredBy", "IterateCells", "Release", "ReleaseCell", "SetAutoHideDelay", "SetCell", "SetCellColor", "SetCellMarginH", "SetCellMarginV", "SetCellScript", "SetCellTextColor", "SetColumnColor", "SetColumnLayout", "SetColumnScript", "SetColumnTextColor", "SetDefaultProvider", "SetFont", "SetHeaderFont", "SetHighlightTexCoord", "SetHighlightTexture", "SetLineColor", "SetLineScript", "SetLineTextColor", "SetScript", "SetScrollStep", "SetupCell", "SmartAnchorTo", "UpdateScrolling"} do
		SixTooltip[key] = function (self, ...)
			if self.tooltip then
				return self.tooltip[key](self.tooltip, ...)
			end
		end
	end

	-- SixTooltip
	Mixin(SixTooltip, SixTooltipMixin)
	SixTooltip.config = CreateFromMixins(CONFIG_DEFAULTS, config)
	SixTooltip.dynamic = {}

	return SixTooltip
end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

function Library:CreateTooltip(config)
	return CreateInstance(config)
end
